@@ -22,7 +22,6 @@ import (
 	"fmt"
 	"reflect"
 	"strings"
-	"time"
 
 	"go.uber.org/zap"
 
@@ -38,7 +37,8 @@ import (
 	taskrunreconciler "github.com/tektoncd/pipeline/pkg/client/injection/reconciler/pipeline/v1beta1/taskrun"
 	listers "github.com/tektoncd/pipeline/pkg/client/listers/pipeline/v1beta1"
 	resourcelisters "github.com/tektoncd/pipeline/pkg/client/resource/listers/resource/v1alpha1"
-	"github.com/tektoncd/pipeline/pkg/contexts"
+	"github.com/tektoncd/pipeline/pkg/clock"
+	"github.com/tektoncd/pipeline/pkg/internal/affinityassistant"
 	"github.com/tektoncd/pipeline/pkg/internal/limitrange"
 	podconvert "github.com/tektoncd/pipeline/pkg/pod"
 	tknreconciler "github.com/tektoncd/pipeline/pkg/reconciler"
@@ -55,6 +55,7 @@ import (
 	"k8s.io/client-go/kubernetes"
 	corev1Listers "k8s.io/client-go/listers/core/v1"
 	"knative.dev/pkg/apis"
+	"knative.dev/pkg/changeset"
 	"knative.dev/pkg/controller"
 	"knative.dev/pkg/kmeta"
 	"knative.dev/pkg/logging"
@@ -66,32 +67,21 @@ type Reconciler struct {
 	KubeClientSet     kubernetes.Interface
 	PipelineClientSet clientset.Interface
 	Images            pipeline.Images
+	Clock             clock.Clock
 
 	// listers index properties about resources
-	taskRunLister     listers.TaskRunLister
-	taskLister        listers.TaskLister
-	clusterTaskLister listers.ClusterTaskLister
-	resourceLister    resourcelisters.PipelineResourceLister
-	limitrangeLister  corev1Listers.LimitRangeLister
-	cloudEventClient  cloudevent.CEClient
-	entrypointCache   podconvert.EntrypointCache
-	metrics           *taskrunmetrics.Recorder
-	pvcHandler        volumeclaim.PvcHandler
-
-	// disableResolution is a flag to the reconciler that it should
-	// not be performing resolution of taskRefs.
-	// TODO(sbwsg): Once we've agreed on a way forward for TEP-0060
-	// this can be removed in favor of whatever that chosen solution
-	// is.
-	disableResolution bool
+	taskRunLister    listers.TaskRunLister
+	resourceLister   resourcelisters.PipelineResourceLister
+	limitrangeLister corev1Listers.LimitRangeLister
+	cloudEventClient cloudevent.CEClient
+	entrypointCache  podconvert.EntrypointCache
+	metrics          *taskrunmetrics.Recorder
+	pvcHandler       volumeclaim.PvcHandler
 }
 
 // Check that our Reconciler implements taskrunreconciler.Interface
 var (
 	_ taskrunreconciler.Interface = (*Reconciler)(nil)
-
-	// Indicates taskrun resolution hasn't occurred yet.
-	errResourceNotResolved = fmt.Errorf("task ref has not been resolved")
 )
 
 // ReconcileKind compares the actual state with the desired, and attempts to
@@ -128,10 +118,10 @@ func (c *Reconciler) ReconcileKind(ctx context.Context, tr *v1beta1.TaskRun) pkg
 
 		// We may be reading a version of the object that was stored at an older version
 		// and may not have had all of the assumed default specified.
-		tr.SetDefaults(contexts.WithUpgradeViaDefaulting(ctx))
+		tr.SetDefaults(ctx)
 
 		// Try to send cloud events first
-		cloudEventErr := cloudevent.SendCloudEvents(tr, c.cloudEventClient, logger)
+		cloudEventErr := cloudevent.SendCloudEvents(tr, c.cloudEventClient, logger, c.Clock)
 		// Regardless of `err`, we must write back any status update that may have
 		// been generated by `sendCloudEvents`
 		if cloudEventErr != nil {
@@ -140,27 +130,10 @@ func (c *Reconciler) ReconcileKind(ctx context.Context, tr *v1beta1.TaskRun) pkg
 			return cloudEventErr
 		}
 
-		pod, err := c.stopSidecars(ctx, tr)
-		if err != nil {
+		if err := c.stopSidecars(ctx, tr); err != nil {
 			return err
 		}
 
-		go func(metrics *taskrunmetrics.Recorder) {
-			err := metrics.DurationAndCount(tr)
-			if err != nil {
-				logger.Warnf("Failed to log the metrics : %v", err)
-			}
-			if pod != nil {
-				err = metrics.RecordPodLatency(pod, tr)
-				if err != nil {
-					logger.Warnf("Failed to log the metrics : %v", err)
-				}
-			}
-			err = metrics.CloudEvents(tr)
-			if err != nil {
-				logger.Warnf("Failed to log the metrics : %v", err)
-			}
-		}(c.metrics)
 		return c.finishReconcileUpdateEmitEvents(ctx, tr, before, nil)
 	}
 
@@ -173,55 +146,74 @@ func (c *Reconciler) ReconcileKind(ctx context.Context, tr *v1beta1.TaskRun) pkg
 
 	// Check if the TaskRun has timed out; if it is, this will set its status
 	// accordingly.
-	if tr.HasTimedOut(ctx) {
+	if tr.HasTimedOut(ctx, c.Clock) {
 		message := fmt.Sprintf("TaskRun %q failed to finish within %q", tr.Name, tr.GetTimeout(ctx))
 		err := c.failTaskRun(ctx, tr, v1beta1.TaskRunReasonTimedOut, message)
 		return c.finishReconcileUpdateEmitEvents(ctx, tr, before, err)
 	}
 
 	// prepare fetches all required resources, validates them together with the
-	// taskrun, runs API convertions. Errors that come out of prepare are
-	// permanent one, so in case of error we update, emit events and return
+	// taskrun, runs API conversions. In case of error we update, emit events and return.
 	_, rtr, err := c.prepare(ctx, tr)
-	if c.disableResolution && err == errResourceNotResolved {
-		// This is not an error - the taskrun is still expected
-		// to be resolved out-of-band.
-	} else {
-		if err != nil {
-			logger.Errorf("TaskRun prepare error: %v", err.Error())
-			// We only return an error if update failed, otherwise we don't want to
-			// reconcile an invalid TaskRun anymore
-			return c.finishReconcileUpdateEmitEvents(ctx, tr, nil, err)
-		}
+	if err != nil {
+		logger.Errorf("TaskRun prepare error: %v", err.Error())
+		// We only return an error if update failed, otherwise we don't want to
+		// reconcile an invalid TaskRun anymore
+		return c.finishReconcileUpdateEmitEvents(ctx, tr, nil, err)
+	}
 
-		// Store the condition before reconcile
-		before = tr.Status.GetCondition(apis.ConditionSucceeded)
+	// Store the condition before reconcile
+	before = tr.Status.GetCondition(apis.ConditionSucceeded)
 
-		// Reconcile this copy of the task run and then write back any status
-		// updates regardless of whether the reconciliation errored out.
-		if err = c.reconcile(ctx, tr, rtr); err != nil {
-			logger.Errorf("Reconcile: %v", err.Error())
-		}
+	// Reconcile this copy of the task run and then write back any status
+	// updates regardless of whether the reconciliation errored out.
+	if err = c.reconcile(ctx, tr, rtr); err != nil {
+		logger.Errorf("Reconcile: %v", err.Error())
+	}
 
-		// Emit events (only when ConditionSucceeded was changed)
-		if err = c.finishReconcileUpdateEmitEvents(ctx, tr, before, err); err != nil {
-			return err
-		}
+	// Emit events (only when ConditionSucceeded was changed)
+	if err = c.finishReconcileUpdateEmitEvents(ctx, tr, before, err); err != nil {
+		return err
 	}
 
 	if tr.Status.StartTime != nil {
 		// Compute the time since the task started.
-		elapsed := time.Since(tr.Status.StartTime.Time)
+		elapsed := c.Clock.Since(tr.Status.StartTime.Time)
 		// Snooze this resource until the timeout has elapsed.
 		return controller.NewRequeueAfter(tr.GetTimeout(ctx) - elapsed)
 	}
 	return nil
 }
-func (c *Reconciler) stopSidecars(ctx context.Context, tr *v1beta1.TaskRun) (*corev1.Pod, error) {
+
+func (c *Reconciler) durationAndCountMetrics(ctx context.Context, tr *v1beta1.TaskRun) {
+	logger := logging.FromContext(ctx)
+	if tr.IsDone() {
+		newTr, err := c.taskRunLister.TaskRuns(tr.Namespace).Get(tr.Name)
+		if err != nil {
+			logger.Errorf("Error getting TaskRun %s when updating metrics: %w", tr.Name, err)
+		}
+		before := newTr.Status.GetCondition(apis.ConditionSucceeded)
+		go func(metrics *taskrunmetrics.Recorder) {
+			if err := metrics.DurationAndCount(tr, before); err != nil {
+				logger.Warnf("Failed to log the metrics : %v", err)
+			}
+			if err := metrics.CloudEvents(tr); err != nil {
+				logger.Warnf("Failed to log the metrics : %v", err)
+			}
+		}(c.metrics)
+	}
+}
+
+func (c *Reconciler) stopSidecars(ctx context.Context, tr *v1beta1.TaskRun) error {
 	logger := logging.FromContext(ctx)
 	// do not continue without knowing the associated pod
 	if tr.Status.PodName == "" {
-		return nil, nil
+		return nil
+	}
+
+	// do not continue if the TaskSpec had no sidecars
+	if tr.Status.TaskSpec != nil && len(tr.Status.TaskSpec.Sidecars) == 0 {
+		return nil
 	}
 
 	// do not continue if the TaskRun was canceled or timed out as this caused the pod to be deleted in failTaskRun
@@ -229,10 +221,15 @@ func (c *Reconciler) stopSidecars(ctx context.Context, tr *v1beta1.TaskRun) (*co
 	if condition != nil {
 		reason := v1beta1.TaskRunReason(condition.Reason)
 		if reason == v1beta1.TaskRunReasonCancelled || reason == v1beta1.TaskRunReasonTimedOut {
-			return nil, nil
+			return nil
 		}
 	}
 
+	// do not continue if there are no Running sidecars.
+	if !podconvert.IsSidecarStatusRunning(tr) {
+		return nil
+	}
+
 	pod, err := podconvert.StopSidecars(ctx, c.Images.NopImage, c.KubeClientSet, tr.Namespace, tr.Status.PodName)
 	if err == nil {
 		// Check if any SidecarStatuses are still shown as Running after stopping
@@ -241,16 +238,15 @@ func (c *Reconciler) stopSidecars(ctx context.Context, tr *v1beta1.TaskRun) (*co
 			err = updateStoppedSidecarStatus(ctx, pod, tr, c)
 		}
 	}
-
 	if k8serrors.IsNotFound(err) {
 		// At this stage the TaskRun has been completed if the pod is not found, it won't come back,
 		// it has probably evicted. We can return the error, but we consider it a permanent one.
-		return nil, controller.NewPermanentError(err)
+		return controller.NewPermanentError(err)
 	} else if err != nil {
 		logger.Errorf("Error stopping sidecars for TaskRun %q: %v", tr.Name, err)
 		tr.Status.MarkResourceFailed(podconvert.ReasonFailedResolution, err)
 	}
-	return pod, nil
+	return nil
 }
 
 func (c *Reconciler) finishReconcileUpdateEmitEvents(ctx context.Context, tr *v1beta1.TaskRun, beforeCondition *apis.Condition, previousError error) error {
@@ -278,37 +274,27 @@ func (c *Reconciler) finishReconcileUpdateEmitEvents(ctx context.Context, tr *v1
 // It may report errors back to Reconcile, it updates the taskrun status in case of
 // error but it does not sync updates back to etcd. It does not emit events.
 // All errors returned by `prepare` are always handled by `Reconcile`, so they don't cause
-// the key to be re-queued directly. Once we start using `PermanentErrors` code in
-// `prepare` will be able to control which error is handled by `Reconcile` and which is not
-// See https://github.com/tektoncd/pipeline/issues/2474 for details.
+// the key to be re-queued directly.
 // `prepare` returns spec and resources. In future we might store
 // them in the TaskRun.Status so we don't need to re-run `prepare` at every
 // reconcile (see https://github.com/tektoncd/pipeline/issues/2473).
 func (c *Reconciler) prepare(ctx context.Context, tr *v1beta1.TaskRun) (*v1beta1.TaskSpec, *resources.ResolvedTaskResources, error) {
 	logger := logging.FromContext(ctx)
-	// We may be reading a version of the object that was stored at an older version
-	// and may not have had all of the assumed default specified.
-	tr.SetDefaults(contexts.WithUpgradeViaDefaulting(ctx))
-
-	if c.disableResolution && tr.Status.TaskSpec == nil {
-		return nil, nil, errResourceNotResolved
-	}
+	tr.SetDefaults(ctx)
 
 	getTaskfunc, err := resources.GetTaskFuncFromTaskRun(ctx, c.KubeClientSet, c.PipelineClientSet, tr)
 	if err != nil {
 		logger.Errorf("Failed to fetch task reference %s: %v", tr.Spec.TaskRef.Name, err)
-		tr.Status.SetCondition(&apis.Condition{
-			Type:    apis.ConditionSucceeded,
-			Status:  corev1.ConditionFalse,
-			Reason:  podconvert.ReasonFailedResolution,
-			Message: err.Error(),
-		})
+		tr.Status.MarkResourceFailed(podconvert.ReasonFailedResolution, err)
 		return nil, nil, err
 	}
 
 	taskMeta, taskSpec, err := resources.GetTaskData(ctx, tr, getTaskfunc)
 	if err != nil {
 		logger.Errorf("Failed to determine Task spec to use for taskrun %s: %v", tr.Name, err)
+		if resources.IsGetTaskErrTransient(err) {
+			return nil, nil, err
+		}
 		tr.Status.MarkResourceFailed(podconvert.ReasonFailedResolution, err)
 		return nil, nil, controller.NewPermanentError(err)
 	}
@@ -340,6 +326,12 @@ func (c *Reconciler) prepare(ctx context.Context, tr *v1beta1.TaskRun) (*v1beta1
 		return nil, nil, controller.NewPermanentError(err)
 	}
 
+	if err := validateTaskSpecRequestResources(ctx, taskSpec); err != nil {
+		logger.Errorf("TaskRun %s taskSpec request resources are invalid: %v", tr.Name, err)
+		tr.Status.MarkResourceFailed(podconvert.ReasonFailedValidation, err)
+		return nil, nil, controller.NewPermanentError(err)
+	}
+
 	if err := ValidateResolvedTaskResources(ctx, tr.Spec.Params, rtr); err != nil {
 		logger.Errorf("TaskRun %q resources are invalid: %v", tr.Name, err)
 		tr.Status.MarkResourceFailed(podconvert.ReasonFailedValidation, err)
@@ -382,6 +374,7 @@ func (c *Reconciler) prepare(ctx context.Context, tr *v1beta1.TaskRun) (*v1beta1
 // error but it does not sync updates back to etcd. It does not emit events.
 // `reconcile` consumes spec and resources returned by `prepare`
 func (c *Reconciler) reconcile(ctx context.Context, tr *v1beta1.TaskRun, rtr *resources.ResolvedTaskResources) error {
+	defer c.durationAndCountMetrics(ctx, tr)
 	logger := logging.FromContext(ctx)
 	recorder := controller.GetEventRecorder(ctx)
 	// Get the TaskRun's Pod if it should have one. Otherwise, create the Pod.
@@ -450,6 +443,9 @@ func (c *Reconciler) reconcile(ctx context.Context, tr *v1beta1.TaskRun, rtr *re
 		if err := podconvert.UpdateReady(ctx, c.KubeClientSet, *pod); err != nil {
 			return err
 		}
+		if err := c.metrics.RecordPodLatency(pod, tr); err != nil {
+			logger.Warnf("Failed to log the metrics : %v", err)
+		}
 	}
 
 	// Convert the Pod's status to the equivalent TaskRun Status.
@@ -498,6 +494,16 @@ func (c *Reconciler) updateTaskRunWithDefaultWorkspaces(ctx context.Context, tr
 }
 
 func (c *Reconciler) updateLabelsAndAnnotations(ctx context.Context, tr *v1beta1.TaskRun) (*v1beta1.TaskRun, error) {
+	// Ensure the TaskRun is properly decorated with the version of the Tekton controller processing it.
+	version, err := changeset.Get()
+	if err != nil {
+		return nil, err
+	}
+	if tr.Annotations == nil {
+		tr.Annotations = make(map[string]string, 1)
+	}
+	tr.Annotations[podconvert.ReleaseAnnotation] = version
+
 	newTr, err := c.taskRunLister.TaskRuns(tr.Namespace).Get(tr.Name)
 	if err != nil {
 		return nil, fmt.Errorf("error getting TaskRun %s when updating labels/annotations: %w", tr.Name, err)
@@ -519,12 +525,7 @@ func (c *Reconciler) handlePodCreationError(ctx context.Context, tr *v1beta1.Tas
 	case isExceededResourceQuotaError(err):
 		// If we are struggling to create the pod, then it hasn't started.
 		tr.Status.StartTime = nil
-		tr.Status.SetCondition(&apis.Condition{
-			Type:    apis.ConditionSucceeded,
-			Status:  corev1.ConditionUnknown,
-			Reason:  podconvert.ReasonExceededResourceQuota,
-			Message: fmt.Sprint("TaskRun Pod exceeded available resources: ", err),
-		})
+		tr.Status.MarkResourceOngoing(podconvert.ReasonExceededResourceQuota, fmt.Sprint("TaskRun Pod exceeded available resources: ", err))
 	case isTaskRunValidationFailed(err):
 		tr.Status.MarkResourceFailed(podconvert.ReasonFailedValidation, err)
 	default:
@@ -553,7 +554,7 @@ func (c *Reconciler) failTaskRun(ctx context.Context, tr *v1beta1.TaskRun, reaso
 	logger.Warnf("stopping task run %q because of %q", tr.Name, reason)
 	tr.Status.MarkResourceFailed(reason, errors.New(message))
 
-	completionTime := metav1.Time{Time: time.Now()}
+	completionTime := metav1.Time{Time: c.Clock.Now()}
 	// update tr completed time
 	tr.Status.CompletionTime = &completionTime
 
@@ -642,7 +643,7 @@ func (c *Reconciler) createPod(ctx context.Context, tr *v1beta1.TaskRun, rtr *re
 	ts = resources.ApplyParameters(ts, tr, defaults...)
 
 	// Apply context substitution from the taskrun
-	ts = resources.ApplyContexts(ts, rtr, tr)
+	ts = resources.ApplyContexts(ts, rtr.TaskName, tr)
 
 	// Apply bound resource substitution from the taskrun.
 	ts = resources.ApplyResources(ts, inputResources, "inputs")
@@ -671,9 +672,6 @@ func (c *Reconciler) createPod(ctx context.Context, tr *v1beta1.TaskRun, rtr *re
 		return nil, err
 	}
 
-	// Check if the HOME env var of every Step should be set to /tekton/home.
-	shouldOverrideHomeEnv := podconvert.ShouldOverrideHomeEnv(ctx)
-
 	// Apply path substitutions for the legacy credentials helper (aka "creds-init")
 	ts = resources.ApplyCredentialsPath(ts, pipeline.CredsDir)
 
@@ -681,9 +679,11 @@ func (c *Reconciler) createPod(ctx context.Context, tr *v1beta1.TaskRun, rtr *re
 		Images:          c.Images,
 		KubeClient:      c.KubeClientSet,
 		EntrypointCache: c.entrypointCache,
-		OverrideHomeEnv: shouldOverrideHomeEnv,
 	}
-	pod, err := podbuilder.Build(ctx, tr, *ts, limitrange.NewTransformer(ctx, tr.Namespace, c.limitrangeLister))
+	pod, err := podbuilder.Build(ctx, tr, *ts,
+		limitrange.NewTransformer(ctx, tr.Namespace, c.limitrangeLister),
+		affinityassistant.NewTransformer(ctx, tr.Annotations),
+	)
 	if err != nil {
 		return nil, fmt.Errorf("translating TaskSpec to Pod: %w", err)
 	}
@@ -697,8 +697,6 @@ func (c *Reconciler) createPod(ctx context.Context, tr *v1beta1.TaskRun, rtr *re
 	return pod, err
 }
 
-type DeletePod func(podName string, options *metav1.DeleteOptions) error
-
 func isExceededResourceQuotaError(err error) bool {
 	return err != nil && k8serrors.IsForbidden(err) && strings.Contains(err.Error(), "exceeded quota")
 }
